;;; cl-pdf copyright (c) 2002 Marc Battyani see license.txt for the details
;;; Proposals by Dmitri Ivanov: divanov_nosp@m_aha.ru (replace _nosp@m_ by @ to reply)

(in-package :pdf)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Encoding

(defclass custom-encoding (encoding)
 ((base-encoding :initarg :base-encoding :reader base-encoding :initform nil)))

(defmethod initialize-instance :after ((encoding custom-encoding) &rest initargs
                                       &key &allow-other-keys)
  (with-slots (base-encoding) encoding
    (when (and base-encoding
               (or (stringp base-encoding) (symbolp base-encoding)))
      (setf base-encoding (gethash base-encoding *encodings* base-encoding)))))

(defun compute-encoding-differences (encoding &optional (from *standard-encoding*))
  (let ((differences (make-array 20 :fill-pointer 0 :adjustable t))
	(range-started nil))
    (flet ((start-range (code)
	     (when (or (and code (not range-started))(and (not code) range-started))
	       (setf range-started code)
	       (when code (vector-push-extend code differences)))))
      (loop with start-code = nil
	    for standard-char-name across (char-names from)
	    for char-name across (char-names encoding)
	    for code from 0
	    do
	    (cond
	      ((and (not char-name) standard-char-name)
	       (start-range code) (vector-push-extend ".notdef" differences))
	      ((and char-name (not (equal char-name standard-char-name)))
	       (start-range code)
	       (vector-push-extend (add-/ char-name) differences))
	      (t (start-range nil)))))
    differences))

(defmethod make-dictionary ((encoding custom-encoding) &key &allow-other-keys)
  (with-slots (base-encoding) encoding
    (make-instance 'dictionary
      :dict-values
      `(("/Type" . "/Encoding")
        ,@(when base-encoding
            `(("/BaseEncoding" . ,(add-/ (name base-encoding)))
              ("/Differences" . ,(compute-encoding-differences encoding
                                                               base-encoding))))))))

(defmethod initialize-instance :after ((encoding-object encoding-object)
                                       &rest init-options
                                       &key encoding &allow-other-keys)
  (setf (content encoding-object) (make-dictionary encoding)))

;;; CAUTION:
;;;  Basing on :win-ansi-encoding fails for embedded Type1 fonts!
;;;  For some installed fonts that are not embedded, :win-ansi-encoding gets better
;;;  results, so it should be specified explicitly for get-font.
;;;  It seems that get-font should not have any default for the encoding parameter.

(defvar *win-1251-encoding*
  (make-instance 'custom-encoding
                 :name "Win1251Encoding" :keyword-name :win-1251-encoding
                 :base-encoding :standard-encoding   ;:win-ansi-encoding doesn't work!
		 :standard-encoding nil                 
                 :char-names #(
	nil nil nil nil nil nil nil nil nil nil nil nil nil nil nil nil
	nil nil nil nil nil nil nil nil nil nil nil nil nil nil nil nil
	"space" "exclam" "quotedbl" "numbersign" "dollar" "percent" "ampersand"
	"quotesingle" "parenleft" "parenright" "asterisk" "plus" "comma" "hyphen"
	"period" "slash" "zero" "one" "two" "three" "four" "five" "six" "seven"
	"eight" "nine" "colon" "semicolon" "less" "equal" "greater" "question" "at"
	"A" "B" "C" "D" "E" "F" "G" "H" "I" "J" "K" "L" "M" "N" "O" "P" "Q" "R" "S"
	"T" "U" "V" "W" "X" "Y" "Z" "bracketleft" "backslash" "bracketright"
        "asciicircum" "underscore" "grave"
        "a" "b" "c" "d" "e" "f" "g" "h" "i" "j" "k" "l" "m" "n" "o" "p" "q" "r" "s"
        "t" "u" "v" "w" "x" "y" "z" "braceleft" "bar" "braceright" "asciitilde"
        nil "Djecyrillic" "Gjecyrillic" "quotesinglbase" "gjecyrillic" "quotedblbase"
        "ellipsis" "dagger" "daggerdbl" "Euro" "perthousand" "Ljecyrillic"
        "guilsinglleft" "Njecyrillic" "Kjecyrillic" "Tshecyrillic" "Dzhecyrillic"
        "djecyrillic" "quoteleft" "quoteright" "quotedblleft" "quotedblright" "bullet"
        "endash" "emdash" NIL "trademark" "ljecyrillic" "guilsinglright" "njecyrillic"
        "kjecyrillic" "tshecyrillic" "dzhecyrillic" "space" "Ushortcyrillic"
        "ushortcyrillic" "Jecyrillic" "currency" "Gheupturncyrillic" "brokenbar"
        "section" "Iocyrillic" "copyright" "Ecyrillic" "guillemotleft" "logicalnot"
        "hyphen" "registered" "Yicyrillic" "degree" "plusminus" "Icyrillic" "icyrillic"
        "gheupturncyrillic" "mu" "paragraph" "periodcentered" "iocyrillic" "numero"
        "ecyrillic" "guillemotright" "jecyrillic" "Dzecyrillic" "dzecyrillic" "yicyrillic"
        "Acyrillic" "Becyrillic" "Vecyrillic" "Gecyrillic" "Decyrillic" "Iecyrillic"
        "Zhecyrillic" "Zecyrillic" "Iicyrillic" "Iishortcyrillic" "Kacyrillic"
        "Elcyrillic" "Emcyrillic" "Encyrillic" "Ocyrillic" "Pecyrillic" "Ercyrillic"
        "Escyrillic" "Tecyrillic" "Ucyrillic" "Efcyrillic" "Khacyrillic" "Tsecyrillic"
        "Checyrillic" "Shacyrillic" "Shchacyrillic" "Hardsigncyrillic" "Yericyrillic"
        "Softsigncyrillic" "Ereversedcyrillic" "IUcyrillic" "IAcyrillic"
        "acyrillic" "becyrillic" "vecyrillic" "gecyrillic" "decyrillic" "iecyrillic"
        "zhecyrillic" "zecyrillic" "iicyrillic" "iishortcyrillic" "kacyrillic"
        "elcyrillic" "emcyrillic" "encyrillic" "ocyrillic" "pecyrillic" "ercyrillic"
        "escyrillic" "tecyrillic" "ucyrillic" "efcyrillic" "khacyrillic" "tsecyrillic"
        "checyrillic" "shacyrillic" "shchacyrillic" "hardsigncyrillic" "yericyrillic"
        "softsigncyrillic" "ereversedcyrillic" "iucyrillic" "iacyrillic")))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Fonts

(defvar *embed-fonts* :default
  "t, nil, or :default (methods make-font-dictionary and font-descriptor decide)")
(defvar *compress-fonts* t "nil or decode filter designator")

(defgeneric make-dictionary (thing &key &allow-other-keys))
(defgeneric font-descriptor (font-metrics &key embed errorp))

(defmethod make-dictionary ((fm font-metrics)
                            &key font (encoding (encoding font)) (embed *embed-fonts*)
                            &allow-other-keys)
  (let ((font-descriptor (font-descriptor fm :embed embed :errorp nil)))
    (make-instance 'dictionary :dict-values
      `(("/Type" . "/Font")
        ("/Subtype" . ,(add-/ (font-type fm)))
        ("/BaseFont" . ,(add-/ (font-name fm)))
        ("/Encoding" . ,(if (standard-encoding encoding)
                            (add-/ (name encoding))
                            (find-encoding-object encoding)))
        ,@(when font-descriptor
            `(("/FirstChar" . 0)
              ("/LastChar" . 255)
              ("/Widths" . ,(pdf-widths font))
              ("/FontDescriptor" . ,font-descriptor))) )) ))

(defmethod initialize-instance :after ((font-object font-object) &rest initargs
                                       &key font (embed *embed-fonts*)
                                       &allow-other-keys)
  (setf (content font-object) (make-dictionary (font-metrics font)
                                               :font font  :embed embed)))

(defmethod font-descriptor (font-metrics &key embed (errorp nil))
  (declare (ignore font-metrics embed))
  (if errorp
      (error "Generic fonts do not have descriptors.")
      nil))

(defmethod font-descriptor ((t1fm t1-font-metrics)
                            &key (embed *embed-fonts*) (errorp t))
  (flet ((conv-dim (d) (round (* 1000 d))))
    (make-instance 'indirect-object :content
      (make-instance 'dictionary  ;:obj-number 0 :no-link t
        :dict-values        
	`(("/Type" . "/FontDescriptor")
	  ("/FontName"  . ,(add-/ (font-name t1fm)))
	  ;; 4=Symbolic - contains characters outside the standard Latin character set.
	  ("/Flags" . 4)
	  ("/FontBBox" . ,(map 'vector #'conv-dim (font-bbox t1fm)))
	  ("/ItalicAngle" . ,(conv-dim (italic-angle t1fm)))
	  ("/Ascent" . ,(conv-dim (ascender t1fm)))
	  ("/Descent" . ,(conv-dim (descender t1fm)))
	  ("/CapHeight" . ,(conv-dim (cap-height t1fm)))
	  ("/XHeight" . ,(conv-dim (x-height t1fm)))
	  ("/StemV" . ,10)
          ;; When binary-data is not available, don't embded.
          ,@(when (and embed (binary-data t1fm))
              `(("/FontFile" . ,(make-instance 'indirect-object :content
                                  (make-instance 'pdf-stream
                                    :content (binary-data t1fm)
                                    :no-compression (not *compress-fonts*)
                                    :dict-values `(;("/Type" . "/Pages") ;remove!
                                                   ("/Length1" . ,(length1 t1fm))
                                                   ("/Length2" . ,(length2 t1fm))
                                                   ("/Length3" . ,(length3 t1fm))
                                                   ))))))))))) 

(defun find-font-object (font &key (embed :default))
  (let ((font-object (cdr (assoc font (fonts *document*))))) 
    (unless font-object
      (setf font-object (make-instance 'font-object :font font :embed embed))
      (push (cons font font-object) (fonts *document*)))
    font-object))

(defun add-font-to-page (font &key (embed :default))
  (let ((font-object (cdr (assoc font (fonts *page*)))))
    (unless font-object
      (setf font-object (find-font-object font :embed embed))
      (push (cons font font-object) (fonts *page*))
      (add-dict-value (font-objects *page*) (name font-object) font-object))
    font-object))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Lispworks bivalent streams works only for base-chars, not simple-chars.
;;; Workaround on Windows:
;;;  1) Create page contents as a string of type +page-character-type+.
;;;     If we did not, format to string-output-to-string fails when encounter
;;;	a Unicode character.
;;;  2) Open document file stream with :element-type (unsigned-byte 8) 
;;;  3) Output text-string to the binary file by converting every
;;;	character to byte code manually according to the Windows code page.
;;;     Only two write-object methods seem to need hanlding national characters.
;;; NB: If we wanted to use NL characters in other places, except contents, for example
;;;	object names or the like, we would need to
;;;	- pervasively define (lw:set-default-character-element-type 'simple-char),
;;;	- rewrite other methods as well.

(defconstant +page-character-type+
  #+(and lispworks win32) 'lw:simple-char
  #+(and lispworks (not win32)) 'base-char
  #-lispworks 'character
  "Default character type of page contents")

#+(and lispworks win32)
(defvar *code-page* (ef::ef-coded-character-set win32:*multibyte-code-page-ef*))

(defmacro force-char-code (char-or-code)
 ;;; Provide mapping Unicode chars to code belonging [0-255] range.
  (let ((char (gensym "char")))
    `(let ((,char ,char-or-code))
      (if (characterp ,char) 
          #+(and lispworks win32) (ef:char-external-code ,char *code-page*)
          #-(and lispworks win32) (char-code ,char)
          ,char))))

(defmethod write-object ((obj string) &optional root-level)
  (declare (ignorable root-level))
  #+(and lispworks win32)
  (if (lw:text-string-p obj)		; may include unicode characters
      (loop for char across obj
            if (lw:base-char-p char)
            do (write-char char *pdf-stream*)
            else do (write-byte (ef:char-external-code char *code-page*) *pdf-stream*))
      (write-string obj *pdf-stream*))
  #-(and lispworks win32)
  (write-string obj *pdf-stream*))

(defmethod write-object ((obj pdf-stream) &optional root-level)
  (declare (ignorable root-level))
  (when (and *compress-streams* (not (no-compression obj))
	     (> (length (content obj)) *min-size-for-compression*))
    (setf (content obj) (compress-string (content obj)))
    (let ((filter (get-dict-value obj "/Filter")))
      (if filter
	  (change-dict-value obj "/Filter" (vector "/FlateDecode" filter))
	  (add-dict-value obj "/Filter" "/FlateDecode"))))
  (call-next-method)
  (write-line "stream" *pdf-stream*)
  #+(and pdf-binary lispworks win32)
  (let ((content (content obj)))
    (if (lw:text-string-p content)	; may include Unicode characters
        (loop for char across content
              if (lw:base-char-p char)
              do (write-char char *pdf-stream*)
              else do(write-byte (ef:char-external-code char *code-page*) *pdf-stream*))
        ;      else do (write-char (code-char (ef:char-external-code char *code-page*))
        ;                          *pdf-stream*))
        ;; Unfortunately, either encode-lisp-string or write-sequence doesn't work here!
        ;(write-sequence (ef:encode-lisp-string content win32:*multibyte-code-page-ef*)
        ;                *pdf-stream*)
        (write-sequence content *pdf-stream*)))
  #+(and pdf-binary (not (and lispworks win32)))
  (write-sequence (content obj) *pdf-stream*)
  #-pdf-binary
  (if (stringp (content obj))
      (write-sequence (content obj) *pdf-stream*)
      (loop for c across (content obj) do
	(write-char (code-char c) *pdf-stream*)))
  (write-char #\Newline *pdf-stream*)
  (write-line "endstream" *pdf-stream*))

(defmethod write-document ((filename pathname) &optional (document *document*))
   (with-open-file (stream filename
                           :direction :output :if-exists :supersede
                           :element-type #+pdf-binary '(unsigned-byte 8) 
                                         #-pdf-binary 'base-char
                           :external-format +external-format+)
     (write-document stream document)))

(defmacro with-page ((&rest args) &body body)
  `(let* ((*page* (make-instance 'page ,@args)))
    (setf (content (content-stream *page*))
     (with-output-to-string (*page-stream* nil :element-type +page-character-type+)
       ,@body))
    t))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Convenience methods

(defmethod print-object ((self char-metrics) stream)
  (print-unreadable-object (self stream :type t)
    (format stream "~a" (name self))))

(defmethod print-object ((self encoding) stream)
  (print-unreadable-object (self stream :identity t :type t)
    (format stream "~a" (name self))))

(defmethod print-object ((self font) stream)
  (print-unreadable-object (self stream :identity t :type t)
    (format stream "~a" (name self))))

(defmethod print-object ((self font-metrics) stream)
  (print-unreadable-object (self stream :identity t :type t)
    (format stream "~a" (full-name self))))

#|
(pushnew :pdf-binary *features*)
(lw:removef *features* :pdf-binary)

(setf diff (pdf::compute-encoding-differences pdf::*win-ansi-encoding*))
(length diff)

(pdf:clear-font-cache)
(setq afm-ariac (pdf:read-afm-file (lw:current-pathname "../afm/cyr/ariac__.afm")
                                   'pdf::t1-font-metrics))

(setq afm (pdf:load-t1-font (lw:current-pathname "../afm/cyr/arial.afm")
                            (lw:current-pathname "../afm/cyr/arial.pfb")))
(setf (encoding-scheme afm) :win-1251-encoding)

(setq f (pdf:get-font "ArialCyr"))
(setq fd (pdf::font-descriptor afm))

(setq diff (compute-encoding-differences *win-1251-encoding*))
(setf (slot-value *win-1251-encoding* 'base-encoding) *standard-encoding*)


 |#
